from lolpop.component.data_splitter.base_data_splitter import BaseDataSplitter
from lolpop.utils import common_utils as utils
import pandas as pd 

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class LocalDataSplitter(BaseDataSplitter): 
    
    __REQUIRED_CONF__ = {
        "config" : ["MODEL_TARGET", "DROP_COLUMNS"]
    }

    def split_data(self, data, *args, **kwargs):
        """Split data

        Args:
            data (pd.DataFrame): DataFrame to split

        Returns:
            data_out (dict(pd.DataFrame)): A dictionary of dataframes. Depending on 
                the provided configuration it can contain the following: 
                    X_train = Features of the training dataset  
                    y_train = Labels of the training dataset 
                    X_valid = Features of the validation dataset
                    y_valid = Labels of the validation dataset
                    X_test = Features of the test/holdout dataset
                    y_test = Labels of the validation dataset

        """
        data_out = data.drop(self._get_config("DROP_COLUMNS",[]),axis=1, errors="ignore")

        if self.problem_type == "timeseries": 
            data_out = self._split_timeseries_data(
                data_out, 
                self._get_config("time_index"), 
                self._get_config("model_target"),
                self._get_config("test_size"), 
                self._get_config("validation_size"),
            )
        else: 
            data_out = self._split_data(
                data_out, 
                self._get_config("MODEL_TARGET"), 
                split_column = self._get_config("SPLIT_COLUMN", None),
                split_classes = self._get_config("SPLIT_CLASSES", {}),  
                split_ratio = self._get_config("SPLIT_RATIO", [0.8, 0.2]), 
                sample_num = self._get_config("SAMPLE_NUM", 100000), 
                use_stratified = self._get_config("USE_STRATIFIED", False),
                categorical_columns=self._get_config("CATEGORICAL_COLUMNS", []),
                include_test = self._get_config("INCLUDE_TEST", False),
                )
        for key in data_out.keys():
            self.log("Created %s dataset with %s rows" %(key, data_out.get(key).shape[0]))

        return data_out

    def _split_timeseries_data(self, data, time_index, target, test_size=0, validation_size=0): 
        """Split time series data. 

        Args:
            data (pd.DataFrame): Dataframe to split. 
            time_index (_type_): Column representing the time index
            target (_type_): Column representing the model target
            test_size (int, optional): Size of the test dataset. Defaults to 0.
            validation_size (int, optional): Size of the validation dataset. Defaults to 0.

        Returns:
            data_out (dict(pd.DataFrame)): A dictionary of dataframes. Depending on 
                the provided configuration it can contain the following: 
                    X_train = Features of the training dataset  
                    y_train = Labels of the training dataset 
                    X_valid = Features of the validation dataset
                    y_valid = Labels of the validation dataset
                    X_test = Features of the test/holdout dataset
                    y_test = Labels of the validation dataset
        """
        data_out = {}

        #make sure data is ordered
        data = data.sort_values(time_index).reset_index(drop=True)

        test = None 
        valid = None
        train = data  
        if test_size > 0: 
            test = train[-test_size:].reset_index(drop=True)
            train = train[:-test_size]
        if validation_size > 0: 
            valid = train[-validation_size:].reset_index(drop=True)
            train = train[:-validation_size]
        
        data_out = self._build_split_dfs(train, valid, target, test=test)

        return data_out 


    def _split_data(self, data, target,  split_column=None, split_classes={},  split_ratio=[0.8,0.2], sample_num=100000, use_stratified=False, categorical_columns=[], include_test=False, reset_index=True): 
        """ Function to split data. 
            Supports random splitting, manual splitting, and stratified. 
            Can also include test set. 

        Args:
            data (pd.DataFrame): Dataframe to split
            target (string): The model target, or label 
            split_column (string, optional): The column to use for a manual split. Defaults to None.
            split_classes (dict(string), optional): A dictionary that maps column values to split datasets. I.E.
              Something like {"train" : "train", "valid" : "valid", "test" : "test"}. This allows you to use arbitrary 
              values in your manual split. This is required if split_column is provided. 
            split_ratio (list(float), optional): The percentage of data to put in each split dataset. This should 
             contain either 2 or 3 floats, representing train, valid, test (optional) datasets, respectively. This
              should also add up to 1!  Not used if a manual split is specified. Defaults to [0.8,0.2].
            sample_num (int, optional): The number of rows to include in your dataset. This is the total 
              number across train, valid, and test sets. Defaults to 100000.
            use_stratified (bool, optional): Whether or not to used stratified sampling. When doing classifications problems, 
              some frameworks will error out when classes with no cardinality do not appear in one of the datasets. Statified sampling
              ensures that at least one member of each class will appear in each dataset. Defaults to False.
            categorical_columns (bool, optional): A list of categorical columns to for 
              grouping use when use_stratified=True. 
            include_test (bool, optional): If a holdout dataset should be created. Defaults to False.

        Returns:
            data_out (dict(pd.DataFrame)): A dictionary of dataframes. Depending on 
                the provided configuration it can contain the following: 
                    X_train = Features of the training dataset  
                    y_train = Labels of the training dataset 
                    X_valid = Features of the validation dataset
                    y_valid = Labels of the validation dataset
                    X_test = Features of the test/holdout dataset
                    y_test = Labels of the validation dataset
        """
        data_out = {}
        test = None 

        #if a split_column is provided, then we assume you wanted to handle everything yourself
        if split_column is not None: 
            train = data[data[split_column] == split_classes["train"]]
            valid = data[data[split_column] == split_classes["valid"]]
            if include_test: 
                test = data[data[split_column] == split_classes["test"]]
            data_out = self._build_split_dfs(train, valid, target, split_column, test)

        else: #random sample
            if use_stratified: 
                strat_df = data.groupby(categorical_columns, group_keys=False).apply(lambda x: x.sample(min(len(x), 2)))
            if data.shape[0] > sample_num: 
                data = data.sample(sample_num)

            train = data.sample(frac=split_ratio[0])
            valid = data.drop(train.index)
            
            if include_test: 
                temp_valid = valid.copy()
                valid = temp_valid.sample(frac=(split_ratio[1]/(split_ratio[1]+split_ratio[2])))
                test = temp_valid.drop(valid.index)
            
            if use_stratified: 
                train = train.merge(strat_df, how="outer")
                valid = valid.merge(strat_df, how="outer")
                if include_test:
                    test = test.merge(strat_df, how="outer")
            
            data_out = self._build_split_dfs(train, valid, target, test=test, reset_index=reset_index)

        return data_out 

    def _build_split_dfs(self, train, valid, target,  split_column="SPLIT", test=None, reset_index=True): 
        """Builds the dictionary of split datasets from the individual dataframes. 

        Args:
            train (pd.Dataframe): training dataframe
            valid (pd.DataFrame): validation dataframe
            target (string): the model target, or label
            split_column (str, optional): the column specifying the manual split, if used.  Defaults to "SPLIT".
            test (pd.DataFrame, optional): test dataframe. Defaults to None.

        Returns:
            data_out (dict(pd.DataFrame)): A dictionary of dataframes. Depending on 
                the provided configuration it can contain the following: 
                    X_train = Features of the training dataset  
                    y_train = Labels of the training dataset 
                    X_valid = Features of the validation dataset
                    y_valid = Labels of the validation dataset
                    X_test = Features of the test/holdout dataset
                    y_test = Labels of the validation dataset
        """
        data_out = {
            "X_train" : train.drop([target, split_column], axis=1, errors="ignore"), 
            "X_valid" : valid.drop([target, split_column], axis=1, errors="ignore"), 
            "y_train": train[target], 
            "y_valid": valid[target], 
        }
        #include test set if specified
        if test is not None: 
            data_out["X_test"] = test.drop([target, split_column], axis=1, errors="ignore")
            data_out["y_test"] = test[target]

        if reset_index: 
            for key in data_out.keys(): 
                data_out[key] = data_out[key].reset_index(drop=True)

        return data_out

    def get_train_test_dfs(self, data, combine_xy=True, combine_train_valid=True):
        """Given the dictionary of split data sets, this produces combined training and test dataframes. 

        Args:
            data (dict(pd.Dataframes)): A dictionary of dataframes. Should be the output of split_data
            combine_xy (bool, optional): Whether to combine features with the model label. Defaults to True.
            combine_train_valid (bool, optional): Whether to combine training and validation datasets. Defaults to True.

        Returns:
            (train, test): a duple of dataframes 
        """
        if combine_train_valid: 
            df_X = pd.concat([data["X_train"],data["X_valid"]], ignore_index=True)
            df_y = pd.concat([data["y_train"],data["y_valid"]], ignore_index=True)
        else: 
            df_X = data["X_train"]
            df_y = data["y_train"]

        if combine_xy: 
            train = pd.concat([df_X,df_y],axis=1)
            test = pd.concat([data["X_test"], data["y_test"]],axis=1)
        else: 
            train = (df_X, df_y)
            test = (data["X_test"],data["y_test"])

        return train, test 


from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import pandas as pd
import snowflake.connector as snow_conn
from snowflake.sqlalchemy import URL
from sqlalchemy import create_engine
from tqdm import tqdm


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class SnowflakeDataConnector(BaseDataConnector):

    __REQUIRED_CONF__ = {"config": ["SNOWFLAKE_ACCOUNT",
                                    "SNOWFLAKE_USER",
                                    "SNOWFLAKE_PASSWORD",
                                    "SNOWFLAKE_DATABASE",
                                    "SNOWFLAKE_SCHEMA",
                                    "SNOWFLAKE_WAREHOUSE"]}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.snowflake_config = utils.load_config([
            "SNOWFLAKE_ACCOUNT",
            "SNOWFLAKE_USER",
            "SNOWFLAKE_PASSWORD",
            "SNOWFLAKE_DATABASE",
            "SNOWFLAKE_SCHEMA",
            "SNOWFLAKE_WAREHOUSE"], self.config)

    def get_data(self, table, sql=None, *args, **kwargs):
        """
        Retrieves data from the SnowflakeDataConnector.

        Args:
            table (str): The name of the table to retrieve data from (default: None).
            sql (str): The SQL command to execute (default: None).

        Returns:
            pd.DataFrame: A pandas DataFrame containing the data retrieved from the Snowflake data warehouse.
        """
        if sql is None:
            if table is not None:
                sql = "SELECT * FROM %s" % table
            else:
                raise Exception(
                    "Table or SQL command must be provided to SnowflakeDataTransformer in order to retrieve data.")

        data = self._load_data(sql, self.snowflake_config)

        return data

    def save_data(self, data, table, *args, **kwargs):
        """
        Saves a pandas DataFrame to the Snowflake data warehouse.

        Args:
            data (pd.DataFrame): The pandas DataFrame to save.
            table (str): The name of the table to save the data to.
        """
        #check if table already exists
        tables = self.get_data(None, sql="SHOW TABLES like '%s'" % table)

        #if table exists, see if we've added new columns or deleted columns
        if tables.shape[0] > 0:
           table_schema = self.get_data(table)
           new_features = [x for x in data.columns if x.upper(
           ) not in table_schema.columns.str.upper()]
           deleted_features = [
               x for x in table_schema.columns if x.upper() not in data.columns.str.upper()]

           #if we have new columns, we need to alter the table definiton
           if len(new_features) > 0:
                sql = "ALTER TABLE %s " % (table)
                for col_name in new_features:
                    col_type = data.dtypes[col_name]
                    sf_type = self.__map_pandas_col_type_to_sf_type(col_type)
                    sql = sql + ("ADD COLUMN %s %s" %
                                 (col_name.upper(), sf_type))
                self.get_data(None, sql=sql)

           #if we've deleted any features, then just add blank rows to the dataframe
           if len(deleted_features) > 0:
               for feature in deleted_features:
                   data[feature] = None

        #now we can save our data
        self._save_data(data, table, self.snowflake_config)

    def __map_pandas_col_type_to_sf_type(self, col_type):
        """
        Maps a pandas DataFrame column type to the corresponding Snowflake data type.

        Args:
            col_type (pd.dtype): The data type of the pandas DataFrame column.

        Returns:
            str: The corresponding Snowflake data type.
        """
        if col_type.kind == 'M':
            # datetime and timestamp columns in pandas are converted to datetime64[ns] dtype, which corresponds to 'datetime64' in Snowflake
            column_type = 'DATETIME'
        elif col_type.kind == 'm':
            # timedelta columns in pandas are converted to timedelta64[ns] dtype, which corresponds to 'variant' in Snowflake
            column_type = 'VARIANT'
        elif col_type.kind == 'b':
            # boolean columns in pandas are converted to 'bool' dtype, which corresponds to 'boolean' in Snowflake
            column_type = 'BOOLEAN'
        elif col_type.kind == 'i' or col_type.kind == 'u':
            # integer columns in pandas are converted to 'int64' dtype, which corresponds to 'integer' in Snowflake
            column_type = 'INTEGER'
        elif col_type.kind == 'f' or col_type.kind == 'c':
            # float columns in pandas are converted to 'float64' dtype, which corresponds to 'float' in Snowflake
            column_type = 'FLOAT'
        else:
            # All other column types in pandas are assumed to be string columns, which correspond to 'varchar' in Snowflake
            column_type = 'VARCHAR'
        return column_type

    #load data into df
    @classmethod
    def _load_data(self, sql, config, *args, **kwargs):
        """
        Loads data from the Snowflake data warehouse and returns it as a pandas DataFrame.

        Args:
            sql (str): The SQL command to execute.
            config: The configuration for the Snowflake data warehouse.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the data retrieved from the Snowflake data warehouse.
        """
        result = pd.DataFrame()
        result = get_from_snowflake(
            sql,
            config.get("SNOWFLAKE_ACCOUNT"),
            config.get("SNOWFLAKE_USER"),
            config.get("SNOWFLAKE_PASSWORD"),
            config.get("SNOWFLAKE_DATABASE"),
            config.get("SNOWFLAKE_SCHEMA"),
            config.get("SNOWFLAKE_WAREHOUSE")
        )
        return result

    def _save_data(self, data, table_name, config, *args, **kwargs):
        """
        Saves a pandas DataFrame to the Snowflake data warehouse.

        Args:
            data (pd.DataFrame): The pandas DataFrame to save.
            table_name (str): The name of the table to save the data to.
            config: The configuration for the Snowflake data warehouse.
        """
        save_to_snowflake(
            data,
            table_name,
            config.get("SNOWFLAKE_ACCOUNT"),
            config.get("SNOWFLAKE_USER"),
            config.get("SNOWFLAKE_PASSWORD"),
            config.get("SNOWFLAKE_DATABASE"),
            config.get("SNOWFLAKE_SCHEMA"),
            config.get("SNOWFLAKE_WAREHOUSE")
        )

#get df from snowflake
def get_from_snowflake(sql, account, user, password, database, schema, warehouse):
    connection = snow_conn.connect(
        account=account,
        user=user,
        password=password,
        database=database,
        schema=schema,
        warehouse=warehouse,
    )
    cur = connection.cursor()
    cur.execute(sql)
    #coerce_float helps in converting numeric types into floats
    #otherwise they end up as strings and mess stuff up
    df = pd.DataFrame.from_records(
        iter(cur), columns=[x[0] for x in cur.description], coerce_float=True)
    
    connection.close()
    
    return df

#save to snowflake


def save_to_snowflake(df, table_name, account, user, password, database, schema, warehouse):
    chunksize = 16384  # maximum allowed by snowflake
    engine = create_engine(URL(
        account=account,
        user=user,
        password=password,
        database=database,
        schema=schema,
        warehouse=warehouse,
    ))
    connection = engine.connect()
    df = df.rename(columns=str.upper)
    with tqdm(total=len(df), desc="Upload to %s.%s.%s" % (database, schema, table_name)) as pbar:
        for i, cdf in enumerate(utils.chunker(df, chunksize)):
            cdf.to_sql(table_name.replace(" ", "_").upper(), con=engine,
                        index=False, if_exists="append", chunksize=chunksize,
                        method=snow_conn.pandas_tools.pd_writer)
            pbar.update(chunksize)
    connection.close()
    engine.dispose()




from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils

from pathlib import Path 
from omegaconf import dictconfig


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class LocalDataTransformer(BaseDataTransformer):

    __REQUIRED_CONF__ = {
        "config": ["transformer_path"]
    }

    __DEFAULT_CONF__ = {
        "config": {"transformer_func": "transform", 
                   "data_connector": "LocalDataConnector"}
    }

    def __init__(self, dependent_integrations=None, *args, **kwargs):
        if dependent_integrations is None: 
            dependent_integrations = {}
            
        super().__init__(dependent_integrations=dependent_integrations, *args, **kwargs)

        transformer_path = Path(self._get_config("transformer_path")) 
        transformer_func = self._get_config("transformer_func")
        
        if transformer_path.exists(): 
            transformer = utils.load_plugin(transformer_path, self)

            self._transform = getattr(transformer, transformer_func)
        else: 
            self.log("Transformer path not found %s" %transformer_path)

        #load data connector
        data_connector_cl_name=self._get_config("data_connector")
        data_connector_config = self._get_config(
            "data_connector_config", {})
        data_connector_cl = utils.load_class(data_connector_cl_name)
        data_connector = data_connector_cl(
            conf = {"config": data_connector_config},
            parent=self,
            is_standalone=True,
            dependent_integrations={"component": {"logger": self.logger, "notifier": self.notifier, "metadata_tracker": self.metadata_tracker}}
            )       
        self.data_connector = data_connector

    def transform(self, input_data, *args, **kwargs):
        """Runs local transformer workflow in python.

        Args:
            input_data (String): Names of file to load and pass to transformer file.

        Returns:
            pd.DataFrame: the transformed data
        """
        if isinstance(input_data,dict) or isinstance(input_data, dictconfig.DictConfig): 
            data = {k: self.data_connector.get_data(v) for k,v in input_data.items()}
        elif isinstance(input_data,str): 
            data = self.data_connector.get_data(input_data)
        else: 
            raise Exception("input_data not a valid type. Expecting dict or str. Found: %s" %str(type(input_data)))
        kwargs = self._get_config("transformer_kwargs",{})

        data_out = self._transform(data, **kwargs)

        return data_out



from lolpop.utils import common_utils as utils
from lolpop.base_integration import BaseIntegration

class BaseRunner(BaseIntegration):

    def __init__(self, *args, **kwargs):
        super().__init__(integration_type="runner", *args, **kwargs)



from lolpop.component.base_component import BaseComponent
from lolpop.utils import common_utils as utils
from typing import Any
from functools import wraps
from hashlib import md5

class BaseCacheManager(BaseComponent):

    __DEFAULT_CONF__ = {
        "config": {
            "decorator_method": "cache_decorator", 
            "integration_types" : ["component"]
        }
    }

    def cache(self, key, value, *args, **kwargs) -> str:
        pass

    def retrieve(self, key, *args, **kwargs) -> Any: 
        pass 

    def equals(self, objA, objB, *args, **kwargs) -> bool: 
        return utils.compare_objects(objA, objB)

    def cache_decorator(self, func, cls): 
        @wraps(func)
        def wrapper(*args, **kwargs):
            obj = args[0] #object should always be first argument
            base_key = self._stringify_input(obj, func, args, kwargs)
            class_name=obj.name
            cache_integration_types = self._get_config("integration_types", ["component"])
            cache_integration_classes = self._get_config("integration_classes", [])
            if cache_integration_classes is not None and len(cache_integration_classes) > 0: 
                apply_cache = class_name in cache_integration_classes
            else: 
                apply_cache = obj.integration_type in cache_integration_types
            obj.log("Checking cache for %s" %base_key, level="DEBUG")
            if (not obj._get_config("skip_cache", False)) and (apply_cache):
                try:
                    skip_func = True
                    # first check args and kwargs to see if there is any difference
                    # between what is provided and what is in the cache
                    i = 1  
                    for arg in args[1:]:  # start at 1 to skip self
                        cache_obj = self.retrieve(f"{base_key}__arg_{str(i)}")
        
                        if not self.equals(cache_obj, arg): 
                            skip_func=False
                            obj.log("Cached object not equivalent for argument %s" %i, level="DEBUG")
                            break 
                        i+=1

                    if skip_func: 
                        for key,val in kwargs.items():
                            cache_obj = self.retrieve(key=f"{base_key}__{key}")
                            if not self.equals(cache_obj,val): 
                                skip_func = False 
                                obj.log("Cached object not equivalent for keyword argument %s" %key, level="DEBUG")
                                break 

                    #also check if the function has changed
                    if skip_func: 
                        cache_obj = self.retrieve(f"{base_key}__{func.__name__}")
                        if not self.equals(cache_obj, func): 
                            skip_func = False 
                            obj.log("Cached function not eqivalent for %s" %func.__name__, level="DEBUG")

                    #and check the component config 
                    if skip_func: 
                        cache_obj = self.retrieve(f"{base_key}__config")
                        if not self.equals(cache_obj, obj.config):
                            skip_func = False
                            obj.log("Cached config not equivalent for %s" %class_name, level="DEBUG")

                    #and finally, check that the output exists
                    if skip_func: 
                        output = self.retrieve(f"{base_key}__output")
                        if output is None: 
                            skip_func = False
                            obj.log("Cached output doesn't exist for %s" %base_key, level="DEBUG")

                    if not skip_func:  
                        obj.log("Insufficient cache hits to skip function %s. Executing %s.%s as normal." %(base_key, class_name, func.__name__), level="DEBUG")

                        #we detected something was different, so run the function and then cache all objects 
                        output = func(*args, **kwargs)
                        #Note: we should only cache new input values after the function successfully runs,
                        # otherwise you will skip it on the next iteration. 

                        i = 1
                        for arg in args[1:]: 
                            self.cache(f"{base_key}__arg_{str(i)}", arg)
                            i+=1

                        for key, value in kwargs.items(): 
                            self.cache(f"{base_key}__{key}", value)

                        self.cache(f"{base_key}__{func.__name__}", func)
                        self.cache(f"{base_key}__config", obj.config)
                        self.cache(f"{base_key}__output", output)

                        obj.log("Successfully cached all inputs and outputs to %s" %base_key, level="DEBUG")

                        return output 
                    else: 
                        obj.log("Detected no changes to input or function for %s Retrieving output from cache." %base_key, level="INFO")
                        return output 
 
                except Exception as e:
                    #prevent unnecessary nesting
                    if str(e).startswith("An error occurred:"):
                        raise e
                    else:
                        raise Exception(f"An error occurred: {e}")
            else: #component has a skip_cache flag, so we'll just short circuit
                obj.log("Cache criteria not met for %s. Running method as normal" % base_key, level="DEBUG")
                return func(*args, **kwargs)
        return wrapper

    #hacky attempt to make inputs map to a unique value
    def _stringify_input(self, obj, func, args, kwargs) -> str: 
        base = f"{obj.name}_{func.__name__}"
        output = base 
        for arg in args: 
            output = f"{output}_a-{utils.convert_arg_to_string(arg)}"
        for k,v in kwargs.items(): 
            output = f"{output}_k-{k}-{utils.convert_arg_to_string(v)}"
        for k,v in obj.config.items(): 
            output = f"{output}_c-{k}-{utils.convert_arg_to_string(v)}"

        #escape slashes
        output = output.replace("/", ".")
        #mac os only allows 256 characters in file names by default
        #lower to 200 to give room for kwarg names
        if len(output)>200:
            output = f"{base}_{md5(output.encode('UTF-8')).hexdigest()}"
        return output 




import typer 
from lolpop.utils import common_utils as utils
from pathlib import Path
import os 
import json
import pandas as pd 
from typer.core import TyperGroup 

class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    

app = typer.Typer(cls = NaturalOrderGroup, help="Upload local data into your data platform. ")


@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))

@app.command("file", help="Seed a file.")
def seed_file(
    source_file: Path = typer.Argument(
        ..., help="Path to the source file"),
    target: str = typer.Argument(
        ..., help="Where the data is going. Should be a table in DW or path in the object store, etc."),
    data_connector_class: str = typer.Option(None, "--data-connector-class", "-c", help="Data Connector class name."),
    data_connector_kwargs: str = typer.Option("{}", "--kwargs", "-k", 
                                          help="Keyword arguments (as a string) to pass into the data connector."),
):
    typer.secho("Loading data connector class %s" % data_connector_class, fg="blue")
    data_connector_cl = utils.load_class(data_connector_class, "component")
    logger_cl = utils.load_class("StdOutLogger", "component")
    data_connector = data_connector_cl(conf={}, 
                                       dependent_integrations={"component": {"logger": logger_cl()}}, 
                                       is_standalone=True,
                                       **json.loads(data_connector_kwargs))
    data_connector.suppress_logger = True
    data_connector.suppress_notifier = True
    typer.secho("Successfully loaded data connector class %s" %
                data_connector_cl, fg="green")

    typer.secho("Begin saving file %s with data connector" %
                source_file, fg="blue")
    data = utils.create_df_from_file(source_file)
    data_connector.save_data(data, target)
    typer.secho("Successfully saved data to %s!" %target, fg="green")


import typer 
from typer.core import TyperGroup
from typing import List
from pathlib import Path 
from lolpop.utils import common_utils as utils 
import json 
class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    
app = typer.Typer(cls = NaturalOrderGroup, help="Utilities for building deployments.")


@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))



@app.command("package", help="package a workflow.")
def package(
    lolpop_class: str = typer.Argument(..., help="Lolpop class to package."),
    config_file: Path = typer.Option(..., "-c", "--config-file",
                                     help="Location of runner configuration file."),
    lolpop_entrypoint: str = typer.Option("build_all", "-e", "--build-method",
                                          help="The method in the runner class to execute."),
    lolpop_module: str = typer.Option(
        "lolpop.runner", "-m", "--module", help="The lolpop module that the lolpop class belongs in."),
    flow_name: str = typer.Option(
        "prefect_entrypoint", "-f", help="The flow name to use in the entrypoint file."),
    packager_class: str = typer.Option(
        None, "--packager", help="The orchestrator class to use to package the workflow."),
    packager_args: List[str] = typer.Option(
        [], help="List of args to pass into the orchestrator class."),
    packager_kwargs: str = typer.Option(
        "{}", help="Dict (as a string) of kwargs to pass into the orchestrator class"),
    package_method: str = typer.Option("package", "-p", help="Package method"),
    package_type: str = typer.Option(
        "docker", "-t", help="Type of package resource to create."),
    packaging_args: List[str] = typer.Option(
        [], help="Arguments to pass into the package_method"),
    packaging_kwargs: str = typer.Option(
        "{}", help="Dict (as a string) of keyword arguments to pass into the package method"),
    local_file: Path = typer.Option(
        None, "-l", "--local-file", help="Local file to use to read class definition instead of reading directly from lolpop. Useful when you want to run a modified local class that isn't properly registered. "),
    skip_validation: bool = typer.Option(
        False, "--skip-validation", help="Skip configuration validation.")
):

    if packager_class is None:
        typer.secho("Packager class is None. Currently lolpo only supports packaging via an orchestrator. Please provide a valid orchestrator calss via '-p <orchestrator_class>'", fg="red")
        raise typer.Exit(code=1)

    plugin_mods = []
    if local_file is not None:
        cl = utils.load_module_from_file(local_file)
        plugin_mods = [cl.__name__]
    typer.secho("Loading class %s" % packager_class, fg="blue")
    packager_cl = utils.load_class(
        packager_class, plugin_mods=plugin_mods, class_type="component")
    if packager_cl is None:
        typer.secho("Failed to properly load class %s. Exiting..." %
                    packager_class, fg="red")
        raise typer.Exit(code=1)
    typer.secho("Loaded %s!" % packager_class, fg="green")

    typer.secho("Initializing class %s with config file %s" %
                (packager_class, config_file), fg="blue")
    packager = packager_cl(conf=config_file, 
                           is_standalone=True,
                           skip_config_validation=skip_validation,
                           *packager_args, **json.loads(packager_kwargs))
    typer.secho("Initialized!", fg="green")

    if hasattr(packager, package_method):
        typer.secho("Executing %s.%s with args %s and kwargs %s" % (
            packager_class, package_method, packaging_args, packaging_kwargs), fg="blue")
        func = getattr(packager, package_method)
        func(lolpop_class=lolpop_class, lolpop_module=lolpop_module, lolpop_entrypoint=lolpop_entrypoint, package_type=package_type,
             *packaging_args, **json.loads(packaging_kwargs))
        typer.secho("Packaging completed!", fg="green")
    else:
        typer.secho("ERROR: Method %s not found in class %s!" %
                    (package_method, packager_class), fg="red")



@app.command("build", help="build a deployment.")
def build(
    deployer_class: str = typer.Option(..., "--deployer", help="The orchestrator class to use to deploy the workflow."),
    config_file: Path = typer.Option(..., "-c", "--config-file",help="Location of runner configuration file."),
    deployer_args: List[str] = typer.Option([], help="List of args to pass into the orchestrator class."),
    deployer_kwargs: str = typer.Option("{}", help="Dict (as a string) of kwargs to pass into the orchestrator class"),
    deployment_method: str = typer.Option("deploy", "-d", help="Deployment method"), 
    deployment_name: str = typer.Option("lolpop-deployment", "-n", help="Name of the deployment."),
    deployment_type: str = typer.Option("docker", "-t", help="Type of deployment to create."),
    deployment_args: List[str] = typer.Option([], help="Arguments to pass into the deployment_method"), 
    deployment_kwargs: str = typer.Option("{}", help="Dict (as a string) of keyword arguments to pass into the deployment method"), 
    local_file: Path = typer.Option(
        None, "-l", "--local-file", help="Local file to use to read class definition instead of reading directly from lolpop. Useful when you want to run a modified local class that isn't properly registered. "),
    skip_validation: bool = typer.Option(False, "--skip-validation", help="Skip configuration validation.")
): 
    
    if deployer_class is None: 
        typer.secho("Deployer class is None. Currently lolpo only supports deployment via an orchestrator. Please provide a valid orchestrator calss via '-d <orchestrator_class>'", fg="red")
        raise typer.Exit(code=1)

    plugin_mods = []
    if local_file is not None: 
        cl = utils.load_module_from_file(local_file)
        plugin_mods = [cl.__name__]
    typer.secho("Loading class %s" %deployer_class, fg="blue")
    deployer_cl = utils.load_class(
        deployer_class, plugin_mods=plugin_mods, class_type="component")
    if deployer_cl is None:
        typer.secho("Failed to properly load class %s. Exiting..." %deployer_class, fg="red")
        raise typer.Exit(code=1)
    typer.secho("Loaded %s!" % deployer_class, fg="green")

    typer.secho("Initializing class %s with config file %s" %(deployer_class, config_file), fg="blue")
    deployer = deployer_cl(conf=config_file, 
                           is_standalone=True,
                           skip_config_validation=skip_validation,
                           *deployer_args, **json.loads(deployer_kwargs))
    typer.secho("Initialized!", fg="green")
    
    if hasattr(deployer, deployment_method):
        typer.secho("Executing %s.%s with args %s and kwargs %s" % (
            deployer_class, deployment_method, deployment_args, deployment_kwargs), fg="blue")
        func = getattr(deployer, deployment_method)
        func(deployment_name=deployment_name, deployment_type=deployment_type, 
             *deployment_args, **json.loads(deployment_kwargs))
        typer.secho("Deployment completed!", fg="green")        
    else: 
        typer.secho("ERROR: Method %s not found in class %s!" %(deployment_method, deployer_class), fg="red")


@app.command("run", help="run a deployment.")
def run(
    deployment_name: str = typer.Argument(..., help="Name of the deployment."),
    deployer_class: str = typer.Option(..., "--deployer",
                                       help="The orchestrator class to use to deploy the workflow."),
    config_file: Path = typer.Option(..., "-c", "--config-file",
                                     help="Location of runner configuration file."),
    deployer_args: List[str] = typer.Option(
        [], help="List of args to pass into the orchestrator class."),
    deployer_kwargs: str = typer.Option(
        "{}", help="Dict (as a string) of kwargs to pass into the orchestrator class"),
    run_method: str = typer.Option(
        "run", "-r", help="Run method"),
    run_args: List[str] = typer.Option(
        [], help="Arguments to pass into the deployment_method"),
    run_kwargs: str = typer.Option(
        "{}", help="Dict (as a string) of keyword arguments to pass into the deployment method"),
    local_file: Path = typer.Option(
        None, "-l", "--local-file", help="Local file to use to read class definition instead of reading directly from lolpop. Useful when you want to run a modified local class that isn't properly registered. "),
    skip_validation: bool = typer.Option(
        False, "--skip-validation", help="Skip configuration validation.")
):

    if deployer_class is None:
        typer.secho("Deployer class is None. Currently lolpo only supports deployment via an orchestrator. Please provide a valid orchestrator calss via '-d <orchestrator_class>'", fg="red")
        raise typer.Exit(code=1)

    plugin_mods = []
    if local_file is not None:
        cl = utils.load_module_from_file(local_file)
        plugin_mods = [cl.__name__]
    typer.secho("Loading class %s" % deployer_class, fg="blue")
    deployer_cl = utils.load_class(
        deployer_class, plugin_mods=plugin_mods, class_type="component")
    if deployer_cl is None:
        typer.secho("Failed to properly load class %s. Exiting..." %
                    deployer_class, fg="red")
        raise typer.Exit(code=1)
    typer.secho("Loaded %s!" % deployer_class, fg="green")

    typer.secho("Initializing class %s with config file %s" %
                (deployer_class, config_file), fg="blue")
    deployer = deployer_cl(conf=config_file, 
                           is_standalone=True, 
                           skip_config_validation=skip_validation,
                           *deployer_args, **json.loads(deployer_kwargs))
    typer.secho("Initialized!", fg="green")

    if hasattr(deployer, run_method):
        typer.secho("Executing %s.%s with args %s and kwargs %s" % (
            deployer_class, run_method, run_args, run_kwargs), fg="blue")
        func = getattr(deployer, run_method)
        run_id, url = func(deployment_name=deployment_name, *run_args, **json.loads(run_kwargs))
        typer.secho("Deployment ran! Created run %s. Visit %s for more information" %(run_id,url), fg="green")
    else:
        typer.secho("ERROR: Method %s not found in class %s!" %
                    (run_method, deployer_class), fg="red")


@app.command("stop", help="stop a deployment.")
def stop(
    deployer_class: str = typer.Option(..., "--deployer",
                                       help="The orchestrator class to use to deploy the workflow."),
    config_file: Path = typer.Option(..., "-c", "--config-file",
                                     help="Location of runner configuration file."),
    deployer_args: List[str] = typer.Option(
        [], help="List of args to pass into the orchestrator class."),
    deployer_kwargs: str = typer.Option(
        "{}", help="Dict (as a string) of kwargs to pass into the orchestrator class"),
    stop_method: str = typer.Option(
        "stop", "-s", help="Stop method"),
    deployment_name: str = typer.Option(
        "lolpop-deployment", "-n", help="Name of the deployment."),
    deployment_type: str = typer.Option(
        "docker", "-t", help="Type of deployment to create."),
    stop_args: List[str] = typer.Option(
        [], help="Arguments to pass into the deployment_method"),
    stop_kwargs: str = typer.Option(
        {}, help="Dict (as a string) of keyword arguments to pass into the deployment method"),
    local_file: Path = typer.Option(
        None, "-l", "--local-file", help="Local file to use to read class definition instead of reading directly from lolpop. Useful when you want to run a modified local class that isn't properly registered. "),
    skip_validation: bool = typer.Option(
        False, "--skip-validation", help="Skip configuration validation.")
):

    if deployer_class is None:
        typer.secho("Deployer class is None. Currently lolpo only supports deployment via an orchestrator. Please provide a valid orchestrator calss via '-d <orchestrator_class>'", fg="red")
        raise typer.Exit(code=1)

    plugin_mods = []
    if local_file is not None:
        cl = utils.load_module_from_file(local_file)
        plugin_mods = [cl.__name__]
    typer.secho("Loading class %s" % deployer_class, fg="blue")
    deployer_cl = utils.load_class(
        deployer_class, plugin_mods=plugin_mods, class_type="component")
    if deployer_cl is None:
        typer.secho("Failed to properly load class %s. Exiting..." %
                    deployer_class, fg="red")
        raise typer.Exit(code=1)
    typer.secho("Loaded %s!" % deployer_class, fg="green")

    typer.secho("Initializing class %s with config file %s" %
                (deployer_class, config_file), fg="blue")
    deployer = deployer_cl(conf=config_file, 
                           is_standalone=True, 
                           skip_config_validation=skip_validation,
                           *deployer_args, **json.loads(deployer_kwargs))
    typer.secho("Initialized!", fg="green")

    if hasattr(deployer, stop_method):
        typer.secho("Executing %s.%s with args %s and kwargs %s" % (
            deployer_class, stop_method, stop_args, stop_kwargs), fg="blue")
        func = getattr(deployer, stop_method)
        func(deployment_name=deployment_name, deployment_type=deployment_type,
             *stop_args, **json.loads(stop_kwargs))
        typer.secho("Finished stop!", fg="green")
    else:
        typer.secho("ERROR: Method %s not found in class %s!" %
                    (stop_method, deployer_class), fg="red")



from lolpop.component.notifier.base_notifier import BaseNotifier
import smtplib
from datetime import datetime


class SMTPNotifier(BaseNotifier):
    __REQUIRED_CONF__ = {
        "config": ["sender_email", "sender_password", "smtp_server", "smtp_port", "recipient_email"]
    }

    def notify(self, msg, level="ERROR", *args, **kwargs):
        """
        Sends an email notification using the provided message and log level.

        Args:
            msg (str): The body of the email message.
            level (str): The severity level of the notification. Default is "ERROR".
            **kwargs: Additional keyword arguments used if needed.

        Returns:
            None
        """
        with smtplib.SMTP(self._get_config("smtp_sever"), self._get_config("smtp_port")) as smtp:
            # Starting TLS encryption
            smtp.starttls()

            # Login using the sender's email address and password
            smtp.login(self._get_config("sender_email"),
                       self._get_config("sender_password"))

            message = """
            To: %s
            From: %s
            Subject: [%s] %s 
            Date: %s 

            %s
            """ % (
                self._get_config("recipient_email"),
                self._get_config("sender_email"),
                level,
                "Notification from lolpop!",
                datetime.utcnow().strftime(" %d/%M/%y %H:%M:%s"),
                msg
            )
            # Sending the message
            smtp.sendmail(self._get_config("sender_email"),
                          self._get_config("recipient_email"), message)



import typer 
import os 
import re 

from lolpop.cli import create, run, test, datagen, seed, extension, deployment
from importlib.metadata import version
from typing import Optional
from pathlib import Path 
from cookiecutter.main import cookiecutter
from lolpop import __template_path__ as lolpop_template_path
from typer.core import TyperGroup 
from importlib.metadata import metadata
from collections import defaultdict
from pprint import pprint

try:
    __version__ = version("lolpop")
except:
    __version__ = "local-dev"

class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()

app = typer.Typer(cls=NaturalOrderGroup,
                  help="lolpop: A software engineering framework for machine learning workflows.",
                  no_args_is_help=True)

app.add_typer(run.app, name="run")
app.add_typer(create.app, name="create")
app.add_typer(datagen.app, name="datagen")
app.add_typer(seed.app, name="seed")
app.add_typer(test.app, name="test")
app.add_typer(deployment.app, name="deployment")
app.add_typer(extension.app, name="extension")

def version_callback(value: bool):
    if value:
        try:
            typer.secho("lolpop version: %s" % __version__, fg="blue")
            raise typer.Exit()
        except Exception as e:
           raise typer.Exit()


@app.callback(context_settings=dict(help_option_names=["-h", "--help"]))
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print lolpop version",
    ),
):
    return True

@app.command("init", help="Initialize a lolpop project.")
def initialize(
    project_name: str = typer.Argument(..., help="Name of the project to create."), 
    project_path: Path = typer.Option(Path(os.getcwd()), help="Path to create project."), 
    template_path: str = typer.Option(
        lolpop_template_path + "/project_template", help="Path to the project template."),
    ): 
    try:
        cookiecutter(template=template_path,
                    extra_context={"project_name": project_name},
                    output_dir=project_path, no_input=True)
        typer.secho("Successfully created lolpop project %s at location %s/%s" %
                    (project_name, project_path, project_name), fg="green")
    except Exception as e: 
        typer.secho("Failed to create project %s: %s" %(project_name, str(e)), fg="red") 

@app.command("help", add_help_option=False, options_metavar="", hidden=True)
def help(ctx: typer.Context):
    """Show CLI usage help."""
    ctx.info_name = None
    typer.echo(ctx.parent.command.get_help(ctx))


@app.command("list-extras", help="list available extra packages")
def list_extras(
    package_name: str = typer.Option("lolpop",
                                       help="Name of package to list available extras for."),
    print_reqs: bool = typer.Option(False, "--print-reqs", help="Print depencencies included in the extra packages.")
):
    extras = metadata(package_name).get_all("Provides-Extra")
    required_dists = metadata(package_name).get_all("Requires-Dist")

    result = defaultdict(list)
    for extra in extras:
        for required_dist in required_dists:
            suffix = f'extra == "{extra}"'
            if suffix in required_dist:
                result[extra].append(re.sub(r";.+", "", required_dist))

    for k, v in result.items():
        typer.secho(k, fg="blue")
        if print_reqs: 
            for req in v: 
                typer.secho(req, fg="yellow")

    typer.secho("\nTo install all extras, you can issue the following command:", fg="green") 
    typer.secho("pip3 install 'lolpop[%s]'" %",".join(result.keys()))


if __name__ == "__main__":
    app()



from lolpop.component.feature_transformer.base_feature_transformer import BaseFeatureTransformer
from lolpop.utils import common_utils as utils

from sklearn import preprocessing
from sklearn.compose import ColumnTransformer

import numpy as np 
import pandas as pd 

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class sklearnFeatureTransformer(BaseFeatureTransformer):
    __REQUIRED_CONF__ = {"config": ["transformers"]
                         }

    __DEFAULT_CONF__ = {
        "config": {"column_transformer_kwargs": {"remainder": "passthrough", "verbose_feature_names_out": False}}
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        #get transformer config and process
        transformers = self._get_config("transformers",[])
        i = 1
        transformer_array = []
        for t in transformers: 
            transformer = self._get_transformer(
                t.get("transformer"),
                *t.get("transformer_args",[]), 
                **t.get("transformer_kwargs",{})
                )
            name = "transform_%s" %i 
            columns = list(t.get("transformer_columns"))
            i+=1 
            transformer_array.append((name, transformer, columns))
        
        #now we can call ColumnTranformer on the transformer_array 
        self.params = self._get_config("column_transformer_kwargs")
        self.transformer = ColumnTransformer(transformers=transformer_array, **self.params)


    def fit(self, X_data, y_data=None, *args, **kwargs):
        """
        Fit the feature transformer to the data.

        Args:
            X_data: The input data to fit the transformer to.
            y_data: The label data. 
            *args: Positional arguments to be passed to the `fit` method.
            **kwargs: Keyword arguments to be passed to the `fit` method.

        Returns:
            ColumnTransformer: The fitted ColumnTransformer object.

        Raises:
            Exception: If no feature transformer is found.

        """
        if not hasattr(self, "transformer"):
            raise Exception(
                "No feature transformer found. Unable to fit transformer.")
        self.transformer = self.transformer.fit(X_data, y_data, **kwargs)

        return self.transformer

    def transform(self, data, *args, **kwargs):
        """
        Apply the fitted feature transformer to transform the input data.

        Args:
            data: The input data to be transformed.
            *args: Positional arguments to be passed to the `transform` method.
            **kwargs: Keyword arguments to be passed to the `transform` method.

        Returns:
            numpy.ndarray or scipy.sparse matrix: The transformed data.

        Raises:
            Exception: If no feature transformer is found.

        """
        if not hasattr(self, "transformer"):
            raise Exception(
                "No feature transformer found. Unable to transform features.")

        #transform data
        transform_out = self.transformer.transform(data, **kwargs)

        #make sure outpt is array-like
        if not isinstance(transform_out, np.ndarray): 
            transform_out = transform_out.toarray() 

        #convert to DF w/ correct column names
        data_out = pd.DataFrame(transform_out, columns=self.transformer.get_feature_names_out())

        return data_out
    
    def fit_transform(self, X_data, y_data=None, *args, **kwargs): 
        """
        Fit the feature transformer to the data and transform it.

        Args:
            X_data: The input data to fit the transformer to and transform.
            y_data: The label data. 
            *args: Positional arguments to be passed to the `fit_transform` method.
            **kwargs: Keyword arguments to be passed to the `fit_transform` method.

        Returns:
            numpy.ndarray or scipy.sparse matrix: The transformed data.

        Raises:
            Exception: If no feature transformer is found.

        """
        if not hasattr(self, "transformer"):
            raise Exception(
                "No feature transformer found. Unable to transform features.")

        data_out = self.transformer.fit_transform(X_data, y_data, **kwargs)

        return data_out

    def _get_transformer(self, transformer_class, *args, **kwargs): 
        """
        Get the scikit-learn transformer object based on the given class name.

        Args:
            transformer_class: The class name of the transformer.
            *args: Positional arguments to be passed to the transformer constructor.
            **kwargs: Keyword arguments to be passed to the transformer constructor.

        Returns:
            object: The instantiated scikit-learn transformer object.

        Raises:
            Exception: If there is an error loading the transformer from `sklearn.preprocessing`.

        """
        try: 
            transformer = getattr(preprocessing, transformer_class)
            return transformer(*args, **kwargs)
        except Exception as e: 
            raise Exception("Error loading transformer %s from sklearn.preprocessing." %transformer_class)


from lolpop.component.resource_version_control.base_resource_version_control import BaseResourceVersionControl
from lolpop.utils import common_utils as utils
from pathlib import Path 
import os 
import joblib
import pandas as pd

# an assumption we make using this component is that dvc is already set up. 
# that is -- `dvc init --subdir` has been run on a subdir of the main git repo
# this is specified by the DVC_DIR variable. And a remote has been set up with 
# the id given in DVC_REMOTE 
@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class dvcVersionControl(BaseResourceVersionControl): 

    __DEFAULT_CONF__ = {
        "config": {"dvc_dir": "dvc/", 
                   "dvc_remote": "local", 
                   "disable_git_commit": False, 
                   "disable_git_push": False, 
                   "git_path_to_dvc_dir": None, 
                   }
    }

    def __init__(self, *args, **kwargs): 
        #set normal config
        super().__init__(*args, **kwargs)
        
        secrets = utils.load_config(["DVC_DIR", "DVC_REMOTE"], 
                                    utils.copy_config_into(self.config,self.__DEFAULT_CONF__.get("config",{})))

        self.dvc_dir = secrets.get("DVC_DIR")
        if self.dvc_dir[-1] != "/":
            self.dvc_dir = self.dvc_dir + "/"
        self.dvc_remote = secrets.get("DVC_REMOTE")

        if self.dvc_remote == "local": 
            output, _ = utils.execute_cmd(["dvc",  "remote", "list"])
            if "local" not in output: 
                local_dir = "%s/dvc" %(self._get_config("local_dir"))
                self.log("Creating dvc remote 'local' for local directory %s" %local_dir)
                _ = utils.execute_cmd(["dvc", "remote", "add", "-d", "local",  local_dir])

        self.git_url = utils.execute_cmd(["git", "config", "--get", "remote.origin.url"], self)[0].strip()
        
        if not self._get_config("disable_git_commit"): 
            #make sure the .dvc/config file is comitted. 
            utils.git_commit_file(".dvc/config", 
                                msg="Committing .dvc/config", 
                                push=False, 
                                path_from_root=self._get_config("git_path_to_dvc_dir"),
                                logger=self)

    def version_data(self, dataset_version, data, key = None, *args, **kwargs): 
        """
        Version the input dataset using dvc (Data Version Control) and output information about the versioned dataset 
        including versioning id and URI.
        
        Args: 
        dataset_version: str, ID of the dataset being versioned.
        data: DataFrame, the dataset to version.
        key: str, optional, default: None, a unique identifier for the dataset version. 
        
        Returns:
        dictionary containing dataset versioning information including the URI and hexsha.
                
        """
        id = self.metadata_tracker.get_resource_id(dataset_version)

        #set up paths
        dvc_path = self.dvc_dir
        data_dir = "%s/data" %dvc_path 
        os.makedirs(data_dir, exist_ok=True)
        if key: 
            dvc_file = "data/%s_%s.csv" %(id, key)
        else: 
            dvc_file = "data/%s.csv" %(id)
        file_path = dvc_path + dvc_file
        
        #dump dataframe to dvc path
        data.to_csv(file_path, index=False)
        
        #commit file to dvc and .dvc file to git, and  then dvc push
        _,_ = utils.execute_cmd(["dvc", "add", file_path], self)
        _,_ = utils.execute_cmd(["dvc",  "commit", file_path], self)
        #git repo needs to have .dvc/config or else dvc doesn't know where to download from when using `dvc get`
        hexsha=None
        if not self._get_config("disable_git_commit"):
            git_push = not self._get_config("disable_git_push")
            hexsha = utils.git_commit_file("%s.dvc" % file_path, 
                                           msg="Committing dvc file", 
                                           push=git_push, 
                                           path_from_root=self._get_config("git_path_to_dvc_dir"),
                                           logger=self)
        _,_ = utils.execute_cmd(["dvc", "push", "--remote", self.dvc_remote], self)
        #I don't think the URI is really needed. We should always access via dvc commands 
        URI, _ = utils.execute_cmd(["dvc", "get", os.getcwd(), file_path, "--show-url"], self)
        
        return {"uri" : URI, "hexsha": hexsha}

    def get_data(self, dataset_version, vc_info=None, key = None, *args, **kwargs):
        """
        Get the versioned dataset from dvc (Data Version Control) using information about the dataset version.

        Args:
        dataset_version: str, ID of the dataset version to retrieve.
        vc_info: dictionary, containing versioning information for the dataset.
        key: str, optional, default: None, a unique identifier for the dataset version.

        Returns:
        DataFrame containing the versioned dataset.

        """
        if vc_info is None: 
            vc_info = self.metadata_tracker.get_vc_info(dataset_version)
        hexsha = vc_info.get("hexsha")
        df = pd.DataFrame()
        if hexsha is not None and hexsha != 'None': 
            id = self.metadata_tracker.get_resource_id(dataset_version)
            if key:
                dvc_file = "data/%s_%s.csv" % (id, key)
            else:
                dvc_file = "data/%s.csv" % (id)

            file_path = self.dvc_dir + dvc_file 
            if self._get_config("git_path_to_dvc_dir") is not None: 
                file_path = str(Path(self._get_config("git_path_to_dvc_dir")) / file_path)
            #if path exists, remove it before downloading. 
            if os.path.exists(dvc_file):
                os.remove(dvc_file)
            _, _ = utils.execute_cmd(["dvc", "get", self.git_url, file_path, "--rev", hexsha, "-o", dvc_file], self)
            df = pd.read_csv(dvc_file)
            
        return df 

    def version_model(self, experiment, model, algo=None, key=None, *args, **kwargs): 
        """
        Version the input model using dvc (Data Version Control) and output information about the versioned model 
        including versioning id and URI.
        
        Args: 
        experiment: str, unique ID of the experiment.
        model: object, the model to version.
        algo: string, the algorithm used to train the model.
        key: str, optional, default:None, a unique identifier for the model version. 
        
        Returns:
        dictionary containing model versioning information including the URI and hexsha.
        
        """
        id = self.metadata_tracker.get_resource_id(experiment)
        model_version_id = self.metadata_tracker.get_parent_id(experiment, type="experiment")
        if key:
            id = "%s_%s" % (id, key)

        # set up directories and dump model
        dvc_path = self.dvc_dir
        model_dir = dvc_path + "models/%s/%s" %(model_version_id, algo)    
        os.makedirs(model_dir, exist_ok=True)
        model_file = "models/%s/%s/%s" %(model_version_id, algo, id)
        model_path = dvc_path + model_file  
        joblib.dump(model, model_path)

        #now commit to dvc
        _,_ = utils.execute_cmd(["dvc", "add", model_path], self)
        _,_ = utils.execute_cmd(["dvc", "commit", model_path], self)
        hexsha=None
        if not self._get_config("disable_git_commit"): 
            git_push = not self._get_config("disable_git_push")
            hexsha = utils.git_commit_file("%s.dvc" % model_path, 
                                           msg="Committing dvc file", 
                                           push = git_push,
                                           path_from_root=self._get_config("git_path_to_dvc_dir"), 
                                           logger=self)
        _,_ = utils.execute_cmd(["dvc", "push", "--remote", self.dvc_remote], self)
        URI, _ = utils.execute_cmd(["dvc", "get", os.getcwd(), model_path, "--show-url"], self)

        return {"uri" : URI, "hexsha": hexsha}
  
    def get_model(self, experiment, key = None, *args, **kwargs): 
        """
        Get the versioned model using versioning information about the model.

        Args:
        experiment: str, unique ID of the experiment.
        key: str, optional, default: None, a unique identifier for the model version.

        Returns:
        the versioned model.

        """
        id = self.metadata_tracker.get_resource_id(experiment)
        if key:
            id = "%s_%s" % (id, key)
        model_version_id = self.metadata_tracker.get_parent_id(experiment, type="experiment")
        algo = self.metadata_tracker.get_metadata(experiment,id="model_trainer")
        hexsha = self.metadata_tracker.get_vc_info(experiment, key="hexsha").get("hexsha")
        model = None
        if hexsha is not None and hexsha != 'None':    
            dvc_file = "models/%s/%s/%s" %(model_version_id,algo,id)
            file_path = self.dvc_dir + dvc_file
            if self._get_config("git_path_to_dvc_dir") is not None:
                file_path = str(Path(self._get_config("git_path_to_dvc_dir")) / file_path)
            #if path exists, remove it before downloading.
            if os.path.exists(dvc_file):
                os.remove(dvc_file)
            _, _ = utils.execute_cmd(["dvc", "get", self.git_url, file_path,  "--rev", hexsha, "-o", dvc_file], self)
            model = joblib.load(dvc_file)
        return model

    def version_feature_transformer(self, experiment, transformer, transformer_class=None, key=None, *args, **kwargs):
        """
        Version the input feature transformer using dvc (Data Version Control) and output 
        information about the versioned transformer including versioning id and URI.
        
        Args: 
        experiment: str, unique ID of the experiment.
        transformer: object, the transform to version.
        transform_class: string, the feature transform class
        key: str, optional, default:None, a unique identifier for the transform version 
        
        Returns:
        dictionary containing transform versioning information including the URI and hexsha.
        
        """
        id = self.metadata_tracker.get_resource_id(experiment)
        model_version_id = self.metadata_tracker.get_parent_id(experiment, type="experiment")
        if key:
            id = "%s_%s" % (id, key)

        # set up directories and dump model
        dvc_path = self.dvc_dir
        transformer_dir = dvc_path + "transformers/%s/%s" % (model_version_id, transformer_class)
        os.makedirs(transformer_dir, exist_ok=True)
        transformer_file = "transformers/%s/%s/%s" % (model_version_id, transformer_class, id)
        transformer_path = dvc_path + transformer_file
        joblib.dump(transformer, transformer_path)

        #now commit to dvc
        _, _ = utils.execute_cmd(["dvc", "add", transformer_path], self)
        _, _ = utils.execute_cmd(["dvc", "commit", transformer_path], self)
        hexsha = None
        if not self._get_config("disable_git_commit"):
            git_push = not self._get_config("disable_git_push")
            hexsha = utils.git_commit_file("%s.dvc" % transformer_path,
                                           msg="Committing dvc file",
                                           push=git_push,
                                           path_from_root=self._get_config("git_path_to_dvc_dir"),
                                           logger=self)
        _, _ = utils.execute_cmd(["dvc", "push", "--remote", self.dvc_remote], self)
        URI, _ = utils.execute_cmd(["dvc", "get", os.getcwd(), transformer_path, "--show-url"], self)

        return {"uri": URI, "hexsha": hexsha}

    def get_feature_transformer(self, experiment, key=None, *args, **kwargs):
        """
        Get the versioned feature transformer using versioning information.

        Args:
        experiment: str, unique ID of the experiment.
        key: str, optional, default: None, a unique identifier for the model version.

        Returns:
        the versioned model.

        """
        id = self.metadata_tracker.get_resource_id(experiment)
        if key:
            id = "%s_%s" % (id, key)
        model_version_id = self.metadata_tracker.get_parent_id(experiment, type="experiment")
        transformer_class = self.metadata_tracker.get_metadata(experiment, id="feature_transformer_class")
        hexsha = self.metadata_tracker.get_vc_info(experiment, key="feature_transformer_hexsha").get("hexsha")
        transformer = None
        if hexsha is not None and hexsha != 'None' and transformer_class is not None:
            dvc_file = "transformers/%s/%s/%s" % (model_version_id, transformer_class, id)
            file_path = self.dvc_dir + dvc_file
            if self._get_config("git_path_to_dvc_dir") is not None:
                file_path = str(Path(self._get_config("git_path_to_dvc_dir")) / file_path)
            #if path exists, remove it before downloading.
            if os.path.exists(dvc_file):
                os.remove(dvc_file)
            _, _ = utils.execute_cmd(
                ["dvc", "get", self.git_url, file_path,  "--rev", hexsha, "-o", dvc_file], self)
            transformer = joblib.load(dvc_file)
        return transformer

    


from lolpop.component.model_repository.base_model_repository import BaseModelRepository
from lolpop.utils import common_utils as utils

from lolpop.component.metadata_tracker.mlflow_metadata_tracker import MLFlowMetadataTracker
from lolpop.utils import mlflow_utils
import mlflow


@utils.decorate_all_methods([utils.error_handler, 
                             utils.log_execution(), 
                             mlflow_utils.check_active_mlflow_run(mlflow)])
class MLFlowModelRepository(BaseModelRepository):
    __REQUIRED_CONF__ = {
        "component": ["metadata_tracker|MLFlowMetadataTracker"],
        "config": []
    }

    def __init__(self, dependent_integrations=None, *args, **kwargs):
        if dependent_integrations is None: 
            dependent_integrations = {}
        #set normal config
        super().__init__(dependent_integrations=dependent_integrations, *args, **kwargs)

        if isinstance(dependent_integrations.get("component",{}).get("metadata_tracker"), MLFlowMetadataTracker):
            self.client = self.metadata_tracker.client
            self.run = self.metadata_tracker.run
            self.url = self.metadata_tracker.url
        else:  # we should only use this if we allow metrics tracking w/o metadata tracking
            tracking_uri = self._get_config("mlflow_tracking_uri")
            experiment_name = self._get_config("mlflow_experiment_name")

            self.client, self.run = mlflow_utils.connect(
                tracking_uri, experiment_name)
            self.url = tracking_uri

            self.log("Using MLFlow in experiment %s with run id: %s" %(experiment_name, self.run.info.run_id), level="INFO")

    def register_model(self, model_version, model, *args, **kwargs):
        """ Registers model  in MLFlow and returns the name of the registered model.
        
        Args:
            model_version (object): the version of the model to register.
            model (object): the  model object to register.
            
        Returns:
            reg_name (str): the name of the registered model.
        """
        model_module = getattr(mlflow, model.mlflow_module.lower()) 

        model_id = self.metadata_tracker.get_resource_id(model_version)
        reg_name = model_id + "_reg"

        #note: if model is already registered, this will simply bump up the verison number
        model_module.log_model(model._get_model(), 
                               artifact_path = "%s_model" %model_id, 
                               registered_model_name=reg_name,
                               **kwargs)

        #log creation of the new model version. unfortunate log_model doesn't give back the right object
        registered_model_version = self.client.get_latest_versions(reg_name, stages=['None'])[0]
        self.log("Successfully logged model %s at version %s" %(reg_name, registered_model_version.version))

        self.metadata_tracker.log_metadata(model_version, id="status", data="REGISTERED")

        return reg_name

    def promote_model(self, registered_model_name, from_stage="None", to_stage="Production", demote_previous_model_versions=True, *args, **kwargs):
        """ Promotes the specified registered model to a given stage and logs the promotion in the metadata tracker.
        
        Args:
            registered_model_name (str): the name of the registered model to promote.
            from_stage (str): the starting stage of the registered model. Default is "None".
            to_stage (str): the target stage to promote the registered model to. Default is "Production".
            demote_previous_model_versions (bool): a boolean variable to enable or disable demotion of previous instances of the registered model. Default is True.
            *args: additional positional arguments.
            **kwargs: additional keyword arguments.
            
        Returns:
            model_version (tuple): a tuple containing the ID of the registered model and an MLFlow run object for the promoted model version.

        """
        registered_model_version = self.client.get_latest_versions(registered_model_name, stages=[from_stage])[0]

        self.client.transition_model_version_stage(
            name=registered_model_name,
            version=registered_model_version.version,
            stage = to_stage, 
            archive_existing_versions=demote_previous_model_versions,
        )

        self.log("Successfully transitioned model %s, version %s to stage %s" %(registered_model_name, registered_model_version.version, to_stage))

        model_version_id = get_model_id_from_reg_model(registered_model_version.name)
        model_status_tag = "%s.status" %model_version_id
        model_version = (model_version_id, self.client.get_run(run_id=registered_model_version.run_id))

        self.metadata_tracker.log_metadata(model_version, id="status", data="PROMOTED")

        if demote_previous_model_versions: 
            prev_version = self.metadata_tracker.get_prev_resource_version(model_version, extra_filters = ["tags.%s = 'DEPLOYED'" %model_status_tag], num=100)
            if prev_version is not None: 
                for run in prev_version[1]: 
                    self.metadata_tracker.log_metadata(model_version, id = "status", data="DEMOTED")

        return model_version

    #approvals are not implemented in mflow, so just return True if called.
    def check_approval(self, promotion, *args, **kwargs):
        return True

    def approve_model(self, promotion, *args, **kwargs):
        return True

def get_model_id_from_reg_model(reg_model): 
    return "_".join(reg_model.split("_")[:-1])


from lolpop.component.data_profiler.base_data_profiler import BaseDataProfiler
from lolpop.utils import common_utils as utils
from ydata_profiling import ProfileReport
from ydata_profiling.visualisation.plot import timeseries_heatmap

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class YDataProfilingDataProfiler(BaseDataProfiler): 
    
    __REQUIRED_CONF__ = {
        "config" : ["local_dir"]
    }


    __DEFAULT_CONF__ = {
        "config": {"YDATA_PROFILE_REPORT_NAME": "YDATA_DATA_PROFILE_REPORT.HTML",
                "YDATA_COMPARISON_REPORT_NAME": "YDATA_DATA_COMPARISON_REPORT.HTML"}
    }

    def profile_data(self, data, *args, **kwargs): 
        """Profiles data using Ydata Profiling

        Args:
            data (pd.DataFrame): A dataframe of the data to profile.  

        Returns:
            data_report (object): Python object of the report 
            file_path (string): file path of the exported report
        """
        is_timeseries = self.problem_type == "timeseries"
        data_report = ProfileReport(data, tsmode=is_timeseries)
        file_path = "%s/%s" % (self._get_config("local_dir"),
                               self._get_config("YDATA_PROFILE_REPORT_NAME"))
        data_report.to_file(file_path)
        
        return data_report, file_path

    def compare_data(self, data, prev_data, *args, **kwargs): 
        """Produces a (probably data drift) report between two data sets 
           using Ydata Profiling. 

        Args:
            data (pd.DataFrame): A dataframe of the "current" data. 
            prev_data (pd.DataFrame): A dataframe of the "historical" data. 

        Returns:
            data_report (object): Python object of the report 
            file_path (string): file path of the exported report
        """
        comparison = None 
        file_path = None 
        (data, prev_data, ok) = utils.compare_data_schemas(self, data, prev_data)
        if ok: 
            is_timeseries = self.problem_type == "timeseries"
            profile = ProfileReport(data, tsmode=is_timeseries)
            old_profile = ProfileReport(prev_data, tsmode=is_timeseries)
            comparison = profile.compare(old_profile)
            file_path = "%s/%s" % (self._get_config("local_dir"), self._get_config("YDATA_COMPARISON_REPORT_NAME"))
            comparison.to_file(file_path)

        return comparison, file_path



from lolpop.component.data_profiler.base_data_profiler import BaseDataProfiler
from lolpop.utils import common_utils as utils
from tslumen import HtmlReport

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class tslumenDataProfiler(BaseDataProfiler):

    __REQUIRED_CONF__ = {
        "config": ["local_dir"]
    }

    __DEFAULT_CONF__ = {
        "config": {"TSLUMEN_PROFILE_REPORT_NAME": "TSLUMEN_DATA_PROFILE_REPORT.HTML"}
    }

    def profile_data(self, data, *args, **kwargs):
        """Profiles data using tslumen

        Args:
            data (pd.DataFrame): A dataframe of the data to profile.  

        Returns:
            data_report (object): Python object of the report 
            file_path (string): file path of the exported report
        """
        ts_data = data.set_index(self._get_config("time_index"))
        if ts_data.index.inferred_freq is None: 
            ts_data = ts_data.asfreq(self._get_config("forecast_frequency"))
        data_report = HtmlReport(ts_data)
        file_path = "%s/%s" %(self._get_config("local_dir"), self._get_config("TSLUMEN_PROFILE_REPORT_NAME"))
        data_report.save(file_path)

        return data_report, file_path


from lolpop.component.model_visualizer.base_model_visualizer import BaseModelVisualizer
from lolpop.utils import common_utils as utils
from yellowbrick.classifier import ClassificationReport, ConfusionMatrix, ROCAUC, PrecisionRecallCurve, DiscriminationThreshold, ClassPredictionError
from yellowbrick.regressor import ResidualsPlot, PredictionError, AlphaSelection, CooksDistance
from yellowbrick.target import ClassBalance, FeatureCorrelation
import os 
from matplotlib import pyplot as plt
import matplotlib 
matplotlib.use('Agg')

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class YellowbrickModelVisualizer(BaseModelVisualizer): 

    __REQUIRED_CONF__ = {
        "component": ["metadata_tracker"]
    }

    def generate_viz(self, data, model, model_version, *args, **kwargs): 
        """Generate visualizations and save plots for model evaluation.

            Args:
            - `data`: A dictionary containing the training and testing data for the model evaluation.
            - `model`: The trained model object for evaluation.
            - `model_version`: The version string for the model being evaluated.

            Returns:
            - Generates several visualization plots as side-effects.
            - No return values.
        """
        
        if self.problem_type == "classification":
            classification_type = utils.get_multiclass(data["y_train"])
            #model = model._get_model() 

            #generate chart for each data split
            #this way is a little inefficient because we have to keep refitting viz
            #but the viz object seems to not allow re-using pyplot. This was the only way 
            #to get all plots to show -- by recreating the object for each split
            for split in set([x.split("_")[1] for x in data.keys()]): 
                #metric report
                viz = ClassificationReport(model)
                self._save_plot(viz, data, split, model_version, "yb_classification_report")

                #confusion matrix
                viz = ConfusionMatrix(model)
                self._save_plot(viz, data, split, model_version, "yb_confusion_matrix")

                #rocauc
                viz = ROCAUC(model)
                self._save_plot(viz, data, split, model_version, "yb_rocauc")

                #comment out for now -- errors if not all classes are represented in the predictions
                ##pr curve
                #viz = PrecisionRecallCurve(model, per_class=True) #per class handles multiclass use cases
                #self._save_plot(viz, data, split, model_version, "yb_prauc")

                #class error
                viz = ClassPredictionError(model)
                self._save_plot(viz, data, split, model_version, "yp_class_error")

                if classification_type == "binary":
                    #discrimination threshold -- only works for binary
                    viz = DiscriminationThreshold(model) 
                    self._save_plot(viz, data, split, model_version, "yb_disctreshold")

            #class balance
            viz = ClassBalance()
            viz.fit(data["y_train"])
            self._save_pyplot("yb_class_balance", "train", model_version)

            #class balance
            viz = FeatureCorrelation(labels=data["X_train"].columns)
            viz.fit(data["X_train"],data["y_train"])
            self._save_pyplot("yb_feature_correlation", "train", model_version)

        elif self.problem_type == "regression": 
            #residuals plot
            for split in set([x.split("_")[1] for x in data.keys()]):
                viz = ResidualsPlot(model)
                self._save_plot(viz, data, split, model_version,"yb_residuals")

                #real vs actual
                viz = PredictionError(model)
                self._save_plot(viz, data, split, model_version, "yb_prediction_error")

                #alpha selection
                #this fails if the model doesn't have a built in CV method, so we'll wrap it. 
                try:
                    viz = AlphaSelection(model)
                    self._save_plot(viz, data, split, model_version, "yb_alpha_selection")
                except: 
                    self.log("Failed to create AlphaSelection plot. It's likely that the ModelTrainer does not contain a built in CV method.", level="WARN") 

                #cooks distance
                viz = CooksDistance()
                viz.fit(data["X_%s" % split], data["y_%s" % split])
                self._save_pyplot("yb_cooks_distance", split, model_version)

            #note: could be fun to add some model selection viz here as well: https://www.scikit-yb.org/en/latest/api/model_selection/index.html
            # and yb has some interesting feature analysis viz that we could use in data processing: https://www.scikit-yb.org/en/latest/api/features/index.html

    def _save_plot(self, viz, data, split, model_version, plot_name): 
        """Fit and score a visualization object and save it to disk.

            Args:
            - `viz`: The Visualization object for plotting.
            - `data`: A dictionary containing the training and testing data for the model evaluation.
            - `split`: The string name of the sub-dataset we're plotting for.
            - `model_version`: The version string for the model being evaluated.
            - `plot_name`: The name of the file to save the plot as.

            Returns:
            - Generates a single visualization plot as a side-effect.
            - No return values.
        """
        viz.fit(data["X_train"], data["y_train"])
        viz.score(data["X_%s" %split], data["y_%s" %split])
        self._save_pyplot(plot_name, split, model_version)

    def _save_pyplot(self, name, label, model_version): 
        """_Save a pyplot plot to disk.

            Args:
            - `name`: The name of the file to save the plot as.
            - `label`: The string label for sub-dataset we're plotting for.
            - `model_version`: The model version object to save the plots to. 

            Returns:
            - Returns a reference to the saved artifact.
            - Saves a single plot as a side-effect.
        """
        local_path = "%s/%s/%s" %(self._get_config("local_dir"), self.metadata_tracker.get_resource_id(model_version), label)
        os.makedirs(local_path, exist_ok=True)
        key = "%s_%s"%(name, label)
        local_file = "%s/%s.png" %(local_path, key)
        plt.savefig(local_file)
        artifact = self.metadata_tracker.log_artifact(model_version, id = key, path = local_file, external=False)
        #plt.clf()
        plt.close()

        return artifact



from lolpop.component.base_component import BaseComponent
from typing import Any 

class BaseDataSynthesizer(BaseComponent):

    def load_data(self, source_file, *args, **kwargs) -> tuple[Any, Any]:
        pass

    def model_data(self, data, *args, **kwargs) -> Any:
        pass

    def sample_data(self, model, num_rows, *args, **kwargs) -> Any:
        pass

    def evaluate_data(self, real_data, synthetic_data, *args, **kwargs) -> list[Any]:
        pass
